package com.flyfish.framework.service;

import com.flyfish.framework.domain.AdminUserDetails;
import com.flyfish.framework.domain.DepartmentQo;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.domain.po.Department;
import com.flyfish.framework.domain.tree.TreeDomain;
import com.flyfish.framework.enums.UserType;
import com.flyfish.framework.service.impl.BaseReactiveServiceImpl;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.BooleanUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.data.mongodb.core.ReactiveMongoOperations;
import org.springframework.data.mongodb.core.aggregation.Aggregation;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.stereotype.Service;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.stream.Collectors;

import static org.apache.commons.lang3.ObjectUtils.defaultIfNull;

/**
 * 部门服务
 *
 * @author wangyu
 */
@Service
@Slf4j
public class DepartmentService extends BaseReactiveServiceImpl<Department> {

    @Resource
    private ReactiveMongoOperations reactiveMongoOperations;

    /**
     * 获取子部门列表
     *
     * @return 结果
     */
    public Mono<Set<String>> getSubCodes(Set<String> parents) {
        Query query = Query.query(Criteria.where("parentIds").in(parents));
        query.fields().include("_id");
        return reactiveMongoOperations.find(query, Department.class).map(Department::getId).collect(Collectors.toSet());
    }

    /**
     * 查询列表
     *
     * @param query 查询
     * @return 结果
     */
    @Override
    public Flux<Department> getList(Qo<Department> query) {
        return convertQo(query).flatMapMany(super::getList);
    }

    /**
     * 在一批编码中找到深度最小的值
     *
     * @param codes 编码们
     * @return 结果
     */
    private Mono<Integer> getMinDepthInIds(Collection<String> codes) {
        return reactiveMongoOperations.aggregate(
                        Aggregation.newAggregation(
                                Aggregation.match(Criteria.where("_id").in(codes)),
                                Aggregation.group().min("depth").as("depth")),
                        Department.class, Department.class)
                .single()
                .map(TreeDomain::getDepth);
    }

    /**
     * 转换qo，经过可见部门转换，更加私密
     *
     * @param query 查询
     * @return 结果
     */
    private Mono<Qo<Department>> convertQo(Qo<Department> query) {
        if (query instanceof DepartmentQo) {
            DepartmentQo qo = (DepartmentQo) query;
            // 如果是非管理员
            if (qo.getUser().getType() != UserType.SUPER_ADMIN) {
                // 占位，包装查询实体
                Mono<Qo<Department>> mono = Mono.just(qo);
                // 使用列表内的部门作为条件
                AdminUserDetails userDetails = (AdminUserDetails) query.getUser();
                // 查询根节点下的节点
                if (Department.ROOT.equals(qo.getParentId())) {
                    qo.setParentId(null);
                    qo.setIds(userDetails.getVisibleDeparts());
                    // 不递归，指定深度，保证单层
                    if (BooleanUtils.isNotTrue(qo.getRecursive())) {
                        // 指定深度，从可见的列表中获取
                        mono = getMinDepthInIds(userDetails.getVisibleDeparts())
                                .map(depth -> {
                                    qo.setDepth(depth);
                                    return qo;
                                });
                    }
                }
                // 父id均为空，返回所有可见部门即可
                if (null == qo.getParentId() && CollectionUtils.isEmpty(qo.getParentIds())) {
                    qo.setIds(userDetails.getVisibleDeparts());
                }
                // 如果存在额外的id，追加记录
                if (StringUtils.isNotBlank(qo.getAdditional())) {
                    Collection<String> ids = defaultIfNull(qo.getIds(), new HashSet<>());
                    ids.add(qo.getAdditional());
                    qo.setIds(ids);
                }
                return mono;
            }
        }
        return Mono.just(query);
    }
}
